"""
Multi object tree updater

The multi object tree updater allows to update a git-tee with a tree that is
defined by object-hashes and their paths. It uses only the minimal necessary
read/write operations to perform this task.

This is mainly used by the object reference system to efficiently store object
references, by using prefixes of the object-hashes as "directory names", i.e. as
tree reference names.
"""
import logging
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import (
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
)

from .gitbackend.subprocess import git_read_tree_node
from .gitbackend.subprocess import git_save_tree_node
from .utils import split_git_lstree_line


logger = logging.getLogger("datalad.metalad.gitmapper.treeupdater")


class EntryType(Enum):
    Directory = 0
    File = 1


@dataclass
class DirEntry:
    type: EntryType
    object_hash: str
    name: str

    def __post_init__(self):
        _check_hash_and_type(self)

    def get_git_tree_entry_elements(self) -> Tuple[str, str, str, str]:
        if self.type == EntryType.Directory:
            flag = "040000"
            entry_type = "tree"
        else:
            flag = "100644"
            entry_type = "blob"
        return flag, entry_type, self.object_hash, self.name


@dataclass
class PathInfo:
    """
    Instances of this class represent object hashes and the prefix_path under which
    those hashes should be, or are, available.
    """
    elements: List[str]
    object_hash: str
    type: EntryType

    def __post_init__(self):
        _check_hash_and_type(self)

    def first(self) -> str:
        return self.elements[0]

    def is_file(self) -> bool:
        return self.is_leaf() and self.type == EntryType.File

    def is_leaf(self) -> bool:
        return len(self.elements) == 1

    def stripped_by(self, amount: int) -> "PathInfo":
        return PathInfo(
            self.elements[amount:],
            self.object_hash,
            self.type)


def _check_hash_and_type(obj: Union[DirEntry, PathInfo]):
    if len(obj.object_hash) != 40:
        raise ValueError(
            f"{type(obj).__name__}: hash format error ({obj.object_hash})")
    if obj.type not in (EntryType.Directory, EntryType.File):
        raise ValueError(
            f"{type(obj).__name__}: unknown type: {obj.type}")


def _check_leaf_or_directory(leaf_infos: List[PathInfo],
                             directory_infos: Dict[str, PathInfo]):
    """Ensure that a leaf is either a file name or a directory name

    While there might only be one file named u, there might be many directories
    named u, but only one type can exist in the new and existing directory info
    entries.
    """
    # Ensure that a name at this level (a path_info.first()) is either a file
    # name or a directory name.
    leaf_and_dir_names = [
        name
        for name in [file_info.first() for file_info in leaf_infos]
        if name in directory_infos.keys()]

    if leaf_and_dir_names:
        raise ValueError("names used for leafs and directories: " + ", ".join(
            leaf_and_dir_names))


def _check_leaf_uniqueness(leaf_infos: List[PathInfo]):
    """ Ensure that a file name is only used once"""
    duplicated_leaf_names = set([
        file_info.first()
        for file_info in leaf_infos
        if [
            file_info.first()
            for file_info in leaf_infos
        ].count(file_info.first()) > 1])

    if duplicated_leaf_names:
        raise ValueError("duplicated file names: " + ", ".join(
            duplicated_leaf_names))


def _check_entries_validity(leaf_infos: List[PathInfo],
                            directory_infos: Dict[str, PathInfo]):

    _check_leaf_or_directory(leaf_infos, directory_infos)
    _check_leaf_uniqueness(leaf_infos)


def _get_dir(repo: Path, object_hash: str) -> List:
    result = []
    for entry in git_read_tree_node(str(repo), object_hash):
        _, object_type, object_hash, name = split_git_lstree_line(entry)
        result.append(
            DirEntry(
                type=EntryType.File
                if object_type == "blob"
                else EntryType.Directory,
                object_hash=object_hash,
                name=name))
    return result


def _write_dir(repo: Path, entries: Iterable[DirEntry]) -> str:
    return git_save_tree_node(
        str(repo),
        [entry.get_git_tree_entry_elements() for entry in entries])


def add_paths(repo: Path,
              path_infos: List[PathInfo],
              root_entries: List[DirEntry]) -> str:
    """
    Add all prefix_path infos to the tree object defined by "root-entries". Load
    subtrees if necessary, write subtrees, when they are completely assembled.

    This method is called recursively, with path_infos shortened by one and
    corresponding root_entries.
    """

    # Build a quickly accessible entry structure
    root_entry_table = {
        root_entry.name: root_entry
        for root_entry in root_entries}

    # Collect all leaf entries, i.e. prefix_path info has only one element.
    leaf_infos = [path_info for path_info in path_infos if path_info.is_leaf()]

    # Collect all directory entries, i.e. for each directory name, get all
    # sub-paths that start with the directory name.
    directory_infos = {
        path_info.first(): [
            matching_path_info.stripped_by(1)
            for matching_path_info in path_infos
            if matching_path_info not in leaf_infos
            and matching_path_info.first() == path_info.first()
        ]
        for path_info in path_infos
        if path_info not in leaf_infos
    }

    _check_entries_validity(leaf_infos, directory_infos)

    resulting_entries = {**root_entry_table}

    # Process leaf entries
    for leaf_info in leaf_infos:

        name = leaf_info.first()

        existing_entry = root_entry_table.get(name, None)
        if existing_entry:
            if existing_entry.type != leaf_info.type:
                raise ValueError(
                    f"cannot convert {existing_entry.type.name} "
                    f"to {leaf_info.type.name}: {name}")
            logger.debug(f"modifying existing entry: {name}")

        resulting_entries[name] = DirEntry(
            type=leaf_info.type,
            object_hash=leaf_info.object_hash,
            name=name)

    # Process non-leaf entries
    for directory_name, sub_path_infos in directory_infos.items():
        if directory_name in root_entry_table:
            if root_entry_table[directory_name].type != EntryType.Directory:
                raise ValueError(
                    f"cannot convert file to directory: {directory_name}")
            directory_hash = root_entry_table[directory_name].object_hash
            sub_dir_entries = _get_dir(repo, directory_hash)
        else:
            sub_dir_entries = []

        written_directory_hash = add_paths(
            repo,
            sub_path_infos,
            sub_dir_entries)

        resulting_entries[directory_name] = DirEntry(
            type=EntryType.Directory,
            object_hash=written_directory_hash,
            name=directory_name)

    return _write_dir(repo, resulting_entries.values())



from typing import (
    cast,
    Any,
    Iterable,
    Optional,
    Tuple
)

from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.mapper.gitmapper.metadatamapper import MetadataGitMapper
from dataladmetadatamodel.mapper.gitmapper.objectreference import (
    add_tree_reference,
    GitReference,
)
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.mtreenode import MTreeNode
from dataladmetadatamodel.mapper.reference import Reference


class MTreeProxy:
    """
    Base class for all classes that use MTreeNode internally,
    but are not themselves mappable. Those classes typically
    require a stored tree, but do not add additional
    "data-fields", but just convenience methods. Examples are
    "DatasetTree" and "FileTree". Still those classes are
    contained within other mappable objects and therefore should
    behave like a mappable object.
    This class proxies the common MappableObject-interface calls
    to the underlying MTreeNode object.
    # TODO: this could probably be a mixin.
    """
    def __init__(self,
                 leaf_class: Any,
                 mtree: Optional[MTreeNode] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(mtree, (type(None), MTreeNode))
        assert isinstance(reference, (type(None), Reference))

        if mtree is None:
            self.mtree = MTreeNode(leaf_class=leaf_class,
                                   realm=realm,
                                   reference=reference)
        else:
            assert mtree.leaf_class == leaf_class
            self.mtree = mtree

    def __contains__(self, path: MetadataPath) -> bool:
        return self.mtree.get_object_at_path(path) is not None

    def ensure_mapped(self) -> bool:
        return self.mtree.ensure_mapped()

    def read_in(self, backend_type="git") -> MappableObject:
        self.mtree.read_in(backend_type)
        return cast(MappableObject, self)

    def write_out(self,
                  destination: Optional[str] = None,
                  backend_type: str = "git",
                  force_write: bool = False) -> Reference:

        # Since a file tree might contain a large number
        # of metadata entries and since each entry requires
        # two blobs to be written, we use a write-cache.

        # TODO: ugly:
        from .filetree import FileTree

        cache_destination = destination or self.mtree.realm
        if isinstance(self, FileTree):
            MetadataGitMapper.cache_realm(cache_destination)

        reference = self.mtree.write_out(destination,
                                         backend_type,
                                         force_write)

        if isinstance(self, FileTree):
            MetadataGitMapper.flush_realm(cache_destination)

        if not reference.is_none_reference():
            add_tree_reference(reference.location)
        return reference

    def purge(self):
        return self.mtree.purge()

    def is_saved_on(self, destination: str):
        return self.mtree.is_saved_on(destination)

    def deepcopy(self,
                 new_mapper_family: Optional[str] = None,
                 new_destination: Optional[str] = None,
                 **kwargs) -> "MTreeProxy":

        return type(self)(self.mtree.deepcopy(new_mapper_family,
                          new_destination,
                          **kwargs))

    def get_paths_recursive(self,
                            show_intermediate: Optional[bool] = False
                            ) -> Iterable[Tuple[MetadataPath, MappableObject]]:
        yield from self.mtree.get_paths_recursive(show_intermediate)



import re
import subprocess
import unittest
from typing import (
    List,
    Tuple,
)

from ..localcache import (
    cache_object,
    get_cache_realm,
)


known_remote_objects = [
    (
        "https://github.com/datalad/datalad.git",
        "b97c71ce4005cd9db0d5f2dda25bdbac303dabe7",
        ["commit .*", "Author: Michael Hanke.*"]
    ),
    (
        "https://github.com/datalad/datalad-metalad.git",
        "9f718f8623e961b93f7f8b6f43f53d94fb405832",
        ["commit .*", "Merge: 9fd8929 898ff01.*"]
    )
]


known_remote_references = [
    (
        "https://github.com/datalad/datalad.git",
        "refs/tags/0.2.1",
        ["tag 0.2.1", "Tagger: Yaroslav Halchenko.*"]
    ),
    (
        "https://github.com/datalad/datalad-metalad.git",
        "refs/tags/0.2.1",
        ["tag 0.2.1", "Tagger: Michael Hanke.*"]
    )
]


class TestLocalCache(unittest.TestCase):

    def cache(self, specs: List[Tuple[str, str, List[str]]]):
        for repo, object_id, _ in specs:
            cache_object(repo, object_id)

    def check_cached(self, specs: List[Tuple[str, str, List[str]]]):
        for repo, object_id, line_patterns in specs:
            result = subprocess.run(
                [
                    "git",
                    "-P",
                    f"--git-dir={get_cache_realm(repo) / '.git'}",
                    "show",
                    object_id
                ],
                stdout=subprocess.PIPE
            )
            self.assertEqual(result.returncode, 0)
            lines = result.stdout.decode().splitlines()
            for index, pattern in enumerate(line_patterns):
                self.assertIsNotNone(re.match(pattern, lines[index]))

    def test_local_cache_object(self):
        self.cache(known_remote_objects)
        self.check_cached(known_remote_objects)

    def test_local_cache_reference(self):
        # check that identical refs in different repos
        # are kept separate
        self.cache(known_remote_references)
        self.check_cached(known_remote_references)



from dataladmetadatamodel.mapper.gitmapper.gitbackend.subprocess import (
    git_load_json,
    git_save_json,
    git_update_ref,
)
from dataladmetadatamodel.mapper.gitmapper.objectreference import GitReference
from dataladmetadatamodel.mapper.mapper import Mapper
from dataladmetadatamodel.mapper.reference import Reference


class VersionListGitMapper(Mapper):
    """
    Map version lists to git objects.
    The objects are blobs containing json strings that
    define a list of primary data-metadata associations.
    """

    def _get_version_records(self,
                             realm: str,
                             reference: Reference) -> dict:
        from dataladmetadatamodel.datasettree import DatasetTree
        from dataladmetadatamodel.metadatapath import MetadataPath
        from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
        from dataladmetadatamodel.versionlist import VersionRecord

        assert isinstance(reference, Reference)

        json_object = git_load_json(realm, reference.location)

        version_records = {}
        for pdm_assoc in json_object:

            reference = Reference.from_json_obj(pdm_assoc["dataset_tree"])
            if reference.class_name in ("DatasetTree", "MTreeNode"):
                mrr_or_tree = DatasetTree(realm=realm, reference=reference)
            elif reference.class_name == "MetadataRootRecord":
                mrr_or_tree = MetadataRootRecord(None, None, None, None, realm, reference)
            else:
                raise ValueError(
                    f"unexpected tree class in primary data-metadata "
                    f"assoc: ({reference.class_name})")

            primary_data_version = pdm_assoc["primary_data_version"]
            prefix_path = MetadataPath(pdm_assoc["path"])
            if primary_data_version not in version_records:
                version_records[primary_data_version] = dict()
            version_records[primary_data_version][prefix_path] = \
                VersionRecord(
                    pdm_assoc["time_stamp"],
                    prefix_path,
                    mrr_or_tree)

        return version_records

    def map_in_impl(self,
                    version_list: "VersionList",
                    realm: str,
                    reference: Reference) -> None:

        from dataladmetadatamodel.versionlist import VersionList

        assert isinstance(version_list, VersionList)
        version_list.version_set = self._get_version_records(realm, reference)

    def map_out_impl(self,
                     version_list: "VersionList",
                     realm: str,
                     force_write: bool
                     ) -> Reference:

        from dataladmetadatamodel.versionlist import VersionList

        assert isinstance(version_list, VersionList)
        json_object = [
            {
                "primary_data_version": primary_data_version,
                "time_stamp": version_record.time_stamp,
                "path": version_record.prefix_path.as_posix(),
                "dataset_tree": version_record.element.write_out(realm).to_json_obj()
            }
            for primary_data_version, prefix_set in version_list.version_set.items()
            for version_record in prefix_set.values()
        ]

        location = git_save_json(realm, json_object)
        return Reference("VersionList", location)


class TreeVersionListGitMapper(VersionListGitMapper):
    def map_in_impl(self,
                    tree_version_list: "TreeVersionList",
                    realm: str,
                    reference: Reference) -> None:

        from dataladmetadatamodel.versionlist import TreeVersionList

        assert isinstance(tree_version_list, TreeVersionList)
        tree_version_list.version_set = self._get_version_records(realm, reference)

    def map_out_impl(self,
                     tree_version_list: "TreeVersionList",
                     realm: str,
                     force_write: bool
                     ) -> Reference:

        from dataladmetadatamodel.versionlist import TreeVersionList

        assert isinstance(tree_version_list, TreeVersionList)
        reference = super().map_out_impl(tree_version_list, realm, force_write)
        git_update_ref(
            realm,
            GitReference.TREE_VERSION_LIST.value,
            reference.location)
        return reference



import time
import unittest
from typing import (
    cast,
    Any,
    List,
    Optional,
    Union,
)
from uuid import UUID

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.mtreeproxy import MTreeProxy
from dataladmetadatamodel.uuidset import UUIDSet
from dataladmetadatamodel.versionlist import VersionList
from dataladmetadatamodel.mapper.reference import Reference


uuid_pattern = "9900{:04x}00000000000000000000{:04x}"
version_pattern = "ea{:04x}000000000000000000000000000000{:04x}"


default_file_tree_paths = [
    MetadataPath("a/b/c"),
    MetadataPath("a/b/a"),
    MetadataPath("b"),
    MetadataPath("c/d/e"),
    MetadataPath("a/x")]

default_dataset_tree_paths = [
    MetadataPath("d1"),
    MetadataPath("d1/d1.1"),
    MetadataPath("d2"),
    MetadataPath("d2/d2.1/d2.1.1"),
    MetadataPath("d3/d3.1")]


def get_uuid(n: int) -> UUID:
    return UUID(uuid_pattern.format(n, n))


def get_location(n: int) -> str:
    return version_pattern.format(n, n)


def get_version(n: int) -> str:
    return get_location(n)


def assert_equal(test_case: unittest.TestCase,
                 object_a: Any,
                 object_b: Any,
                 _: bool,
                 ):
    test_case.assertEqual(object_a, object_b)


def assert_metadata_equal(test_case: unittest.TestCase,
                          object_a: Any,
                          object_b: Any,
                          _: bool,
                          ):

    test_case.assertEqual(object_a, object_b)


def assert_mrr_equal(test_case: unittest.TestCase,
                     a_mrr: MetadataRootRecord,
                     b_mrr: MetadataRootRecord,
                     a_purge_unsafe: bool,
                     ):

    test_case.assertEqual(a_mrr.dataset_identifier, b_mrr.dataset_identifier)
    test_case.assertEqual(a_mrr.dataset_version, b_mrr.dataset_version)
    assert_mappable_objects_equal(
        test_case,
        a_mrr.dataset_level_metadata,
        b_mrr.dataset_level_metadata,
        a_purge_unsafe,
        assert_metadata_equal)

    assert_mappable_objects_equal(
        test_case,
        a_mrr._file_tree,
        b_mrr._file_tree,
        a_purge_unsafe,
        assert_file_trees_equal)


def assert_mappable_objects_equal(test_case: unittest.TestCase,
                                  a_object: Union[MappableObject, MTreeProxy],
                                  b_object: Union[MappableObject, MTreeProxy],
                                  a_purge_unsafe: bool,
                                  equality_asserter):

    a_object = a_object.read_in()
    b_object = b_object.read_in()

    equality_asserter(test_case, a_object, b_object, a_purge_unsafe)

    if a_purge_unsafe is False:
        a_object.purge()
    b_object.purge()


def assert_file_trees_equal(test_case: unittest.TestCase,
                            a: FileTree,
                            b: FileTree,
                            unsafe: bool
                            ):

    a_entries = list(a.get_paths_recursive())
    b_entries = list(b.get_paths_recursive())

    # Compare paths
    test_case.assertListEqual(
        sorted(list(map(lambda x: x[0], a_entries))),
        sorted(list(map(lambda x: x[0], b_entries))))

    # Compare metadata elements
    for a_object, b_object in zip(
            map(lambda x: x[1], a_entries),
            map(lambda x: x[1], b_entries)):

        # TODO: proper metadata comparison instead of assertEqual
        assert_mappable_objects_equal(
            test_case,
            a_object,
            b_object,
            unsafe,
            assert_equal)


def assert_dataset_trees_equal(test_case: unittest.TestCase,
                               a: DatasetTree,
                               b: DatasetTree,
                               a_purge_unsafe: bool
                               ):

    a_entries = list(a.dataset_paths)
    b_entries = list(b.dataset_paths)

    # Compare paths
    test_case.assertListEqual(
        list(map(lambda x: x[0], a_entries)),
        list(map(lambda x: x[0], b_entries)))

    # Compare metadata_root_records
    for a_mrr, b_mrr in zip(
            map(lambda x: x[1], a_entries),
            map(lambda x: x[1], b_entries)):

        assert_mrr_equal(
            test_case,
            a_mrr.read_in(),
            b_mrr.read_in(),
            a_purge_unsafe)


def assert_mrrs_equal(test_case: unittest.TestCase,
                      a: MetadataRootRecord,
                      b: MetadataRootRecord,
                      unsafe: bool
                      ):

    # Compare dataset level metadata
    assert_mappable_objects_equal(
        test_case,
        a.dataset_level_metadata,
        b.dataset_level_metadata,
        unsafe,
        assert_equal)

    # Compare file trees
    assert_mappable_objects_equal(
        test_case,
        a._file_tree,
        b._file_tree,
        unsafe,
        assert_file_trees_equal
    )

    # Compare the remaining metadata
    test_case.assertEqual(a.dataset_identifier, b.dataset_identifier)
    test_case.assertEqual(a.dataset_version, b.dataset_version)


def assert_version_lists_equal(test_case: unittest.TestCase,
                               a: VersionList,
                               b: VersionList,
                               unsafe: bool
                               ):

    a_entries = [
        a.get_versioned_element(pdv, pref)
        for pdv, pref in a.versions_and_prefix_paths()
    ]
    b_entries = [
        b.get_versioned_element(pdv, pref)
        for pdv, pref in b.versions_and_prefix_paths()
    ]

    for a_entry, b_entry in zip(a_entries, b_entries):
        test_case.assertEqual(a_entry[0], b_entry[0])
        test_case.assertEqual(a_entry[1], b_entry[1])
        test_case.assertIsInstance(a_entry[2], (DatasetTree, MetadataRootRecord))
        test_case.assertIsInstance(b_entry[2], (DatasetTree, MetadataRootRecord))
        if isinstance(a_entry[2], DatasetTree):
            assert_dataset_trees_equal(
                test_case,
                cast(DatasetTree, a_entry[2]),
                cast(DatasetTree, b_entry[2]),
                unsafe)
        else:
            assert_mrr_equal(
                test_case,
                cast(MetadataRootRecord, a_entry[2]),
                cast(MetadataRootRecord, b_entry[2]),
                unsafe)


def assert_uuid_sets_equal(test_case: unittest.TestCase,
                           a_uuid_set: UUIDSet,
                           b_uuid_set: UUIDSet):

    for dataset_id in a_uuid_set.uuids():
        a_version_list = a_uuid_set.get_version_list(dataset_id)
        b_version_list = b_uuid_set.get_version_list(dataset_id)
        assert_version_lists_equal(test_case,
                                   a_version_list,
                                   b_version_list,
                                   True)


def create_file_tree(paths: Optional[List[MetadataPath]] = None) -> FileTree:

    paths = paths or default_file_tree_paths

    file_tree = FileTree()
    for path in paths:
        metadata = Metadata()
        file_tree.add_metadata(path, metadata)
    return file_tree


def create_file_tree_with_metadata(paths: List[MetadataPath],
                                   metadata: List[Metadata]) -> FileTree:

    assert len(paths) == len(metadata)

    file_tree = FileTree()
    for path, md in zip(paths, metadata):
        file_tree.add_metadata(path, md)

    return file_tree


def create_dataset_tree(dataset_tree_paths: Optional[List[MetadataPath]] = None,
                        file_tree_paths: Optional[List[MetadataPath]] = None,
                        initial_mrr: Optional[MetadataRootRecord] = None
                        ) -> DatasetTree:

    dataset_tree_paths = dataset_tree_paths or default_dataset_tree_paths
    file_tree_paths = file_tree_paths or default_file_tree_paths

    dataset_tree = DatasetTree()
    for index, path in enumerate(dataset_tree_paths):

        metadata = Metadata()
        file_tree = create_file_tree(file_tree_paths)

        if initial_mrr is None:
            mrr = MetadataRootRecord(
                get_uuid(index),
                get_version(index),
                metadata,
                file_tree)
        else:
            mrr = initial_mrr

        dataset_tree.add_dataset(path, mrr)

    return dataset_tree


class MMDummy:
    def __init__(self,
                 info: str = "",
                 is_mapped: bool = False,
                 copied_from: Optional["MMDummy"] = None):
        self.info = info or f"MMDummy created at {time.time()}"
        self.mapped = is_mapped
        self.copied_from = copied_from

    def deepcopy(self, *args, **kwargs):
        return MMDummy(self.info, False, self)

    def read_in(self, backend_type="git") -> Any:
        self.mapped = True
        return self

    def write_out(self,
                  destination: Optional[str] = None,
                  backend_type: str = "git",
                  force_write: bool = False) -> Reference:
        return Reference(
            "MMDummy",
            get_location(0x51))

    def purge(self, force: bool = False):
        self.mapped = False
        return



import logging


from dataladmetadatamodel.mapper.gitmapper.objectreference import flush_object_references

from tools.metadata_creator.mrrcreator import create_mrrs_from_dataset
from tools.metadata_creator.treeversionlistcreator import create_tree_version_list_for_mrrs
from tools.metadata_creator.uuidsetcreator import create_uuid_set_for_mrrs


mdc_logger = logging.getLogger("metadata_creator")


def create_metadata_from_dataset(mapper: str,
                                 realm: str,
                                 dataset_path: str,
                                 parameter_set_count: int
                                 ):

    metadata_root_records = create_mrrs_from_dataset(
        mapper,
        realm,
        dataset_path,
        parameter_set_count)

    uuid_set = create_uuid_set_for_mrrs(
        mapper,
        realm,
        metadata_root_records)
    mdc_logger.info(f"saving uuid set: {uuid_set}")
    uuid_set.save()
    mdc_logger.info(f"done saving uuid set: {uuid_set}")

    tree_version_list = create_tree_version_list_for_mrrs(
        mapper,
        realm,
        metadata_root_records)
    mdc_logger.info(f"saving tree version list: {tree_version_list}")
    tree_version_list.save()
    mdc_logger.info(f"done saving tree version list: {tree_version_list}")

    flush_object_references(Path(realm))



"""
Simple lock interface.
"""
import os
import subprocess
import time
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Dict,
    Tuple,
)

from fasteners import InterProcessLock

from dataladmetadatamodel.log import logger


PID = os.getpid()

GIT_MAPPER_LOCK_FILE_NAME = "datalad-metalad-write.lock"
read_write_locked = dict()


@dataclass
class LockState:
    counter: int
    lock: InterProcessLock


def _get_lock_state(lock_dict: dict, lock_dir: Path) -> LockState:
    if lock_dir not in lock_dict:
        lock_dict[lock_dir] = LockState(
            0,
            InterProcessLock(str(lock_dir / GIT_MAPPER_LOCK_FILE_NAME)))
    return lock_dict[lock_dir]


@contextmanager
def locked_backend(realm: Path):
    """ lock and unlock the backend given by realm """
    lock_backend(realm)
    try:
        yield None
    finally:
        unlock_backend(realm)


def lock_backend(realm: Path):
    lock_dir = get_lock_dir(realm)
    lock_state = _get_lock_state(read_write_locked, lock_dir)
    if lock_state.counter == 0:
        lock_time = time.time()
        lock_state.lock.acquire()
        lock_time = time.time() - lock_time
        logger.debug(
            "process {} locked git backend using file {} in {} seconds".format(
                PID,
                lock_dir,
                int(lock_time)))
    lock_state.counter += 1


def unlock_backend(realm: Path):
    lock_dir = get_lock_dir(realm, create_directory=False)
    assert lock_dir.exists()
    lock_state = _get_lock_state(read_write_locked, lock_dir)
    assert lock_state.counter > 0
    lock_state.counter -= 1
    if lock_state.counter == 0:
        logger.debug(
            "process {} unlocks git backend using file {}".format(PID, lock_dir))
        lock_state.lock.release()


def get_lock_dir(realm: Path, create_directory: bool = True) -> Path:
    lock_dir = realm / ".git"
    if not lock_dir.exists() and create_directory:
        # if exist_ok is False, we might get an error if
        # a concurrent mapper tries to lock the same
        # dataset.
        os.makedirs(lock_dir, exist_ok=True)
    return lock_dir


def create_file_tree(location: Path, content: Dict):
    for name, value_or_dict in content.items():
        new_location = location / name
        if isinstance(value_or_dict, str):
            new_location.write_text(value_or_dict)
        else:
            new_location.mkdir()
            create_file_tree(new_location, value_or_dict)


def create_git_repo(location: Path, content: Dict):
    subprocess.run(["git", "init", str(location)], check=True)
    create_file_tree(location, content)
    subprocess.run(
        ["git", "-C", str(location), "add", "."],
        check=True
    )

    subprocess.run(
        ["git", "-C", str(location), "commit", "-m", "create repo"],
        check=True
    )

    commit = subprocess.run(
        ["git", "-C", str(location), "cat-file", "-p", "HEAD"],
        check=True,
        stdout=subprocess.PIPE
    )
    return commit.stdout.decode().splitlines()[0].split()[1]


def split_git_lstree_line(line: str) -> Tuple[str, str, str, str]:
    fixed_fields, name = line.split("\t")
    flag, node_type, hash_value = fixed_fields.split(" ")
    return flag, node_type, hash_value, name



from abc import (
    ABCMeta,
    abstractmethod,
)
from contextlib import contextmanager
from typing import (
    Iterable,
    Optional,
)

from dataladmetadatamodel.log import logger
from dataladmetadatamodel.modifiableobject import ModifiableObject
from dataladmetadatamodel.mapper.reference import Reference


@contextmanager
def ensure_mapped(mappable_object):
    needs_purge = False
    try:
        if mappable_object is not None:
            needs_purge = mappable_object.ensure_mapped()
        yield mappable_object
    finally:
        if needs_purge:
            mappable_object.purge()


class MappableObject(ModifiableObject, metaclass=ABCMeta):
    """
    Base class for objects that can be mapped onto a
    storage backend.
    """
    def __init__(self,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):
        """
        Create a mappable object with a reference or None.
        If the reference is given, we assume that the object
        is saved on the respective realm, and that it is
        not modified.

        We also assume that an object that is created with
        a reference is not mapped. That means it has to be
        read in before operating on its elements.

        Generally unmapped objects are expected to be
        unmodified.
        """
        assert isinstance(reference, (type(None), Reference)), \
            f"MappableObject {self} initialized with invalid reference: {reference}"

        # If the reference is given, we need a realm in
        # which the reference is valid.
        assert reference is None or realm is not None, \
            f"reference provided but no realm in {type(self).__name__} " \
            f"construction."

        # We assume that objects that carry a reference have been
        # saved in the location given by the reference.
        super().__init__(
            realm
            if reference is not None and not reference.is_none_reference()
            else None
        )

        self.realm = realm
        self.reference = reference
        self.mapped = reference is None

    @property
    def modifiable_sub_objects(self) -> Iterable["MappableObject"]:
        """
        Mappable objects might be mapped (in memory) or not mapped
        (stored on secondary storage and purged in order to consume
        as little memory as possible).
        If on abject is not mapped, we assume that it is not modified
        because modifying means: "read-in", "modify", "write-out", and
        "purge". Since "purge" will only succeed if the object was
        written out.
        """
        if not self.mapped:
            return []

        # delegate to our subclasses
        return self.modifiable_sub_objects_impl()

    def read_in(self,
                backend_type: str = "git"
                ) -> "MappableObject":

        from dataladmetadatamodel.mapper import get_mapper

        if self.mapped is False:

            assert self.realm is not None
            assert self.reference is not None

            # If the reference is a None-reference,
            # we can handle this here.
            if self.reference.is_none_reference():
                assert self.reference.class_name == type(self).__name__
                logger.warning(f"read_in({self}): None-reference in {self}")
                self.purge_impl()
                return self

            # Ensure that the object is saved on the given realm
            if not self.is_saved_on(self.realm):
                logger.error(
                    f"read_in({self}): trying to overwrite a modified object")
                raise RuntimeError(
                    "read_in({self}): tried to read over a modified object")

            # The object is not mapped, but saved on self.realm,
            # use the mappable object-specific mapper to read
            # the object in.
            get_mapper(
                type(self).__name__,
                backend_type).map_in(
                    self,
                    self.realm,
                    self.reference)

            # Mark the object as mapped.
            self.mapped = True

        else:

            logger.debug(
                f"read_in({self}): not needed, object is already mapped")

        return self

    def write_out(self,
                  destination_realm: Optional[str] = None,
                  backend_type: str = "git",
                  force_write: bool = False) -> Reference:

        from dataladmetadatamodel.mapper import get_mapper

        # If the object is not mapped and not modified,
        # we do not have to do anything.
        if not self.mapped:

            assert isinstance(self.realm, str), \
                f"write_out: object {self} has no valid " \
                f"realm: {self.realm}"
            assert isinstance(self.reference, Reference), \
                f"write_out: object {self} has no valid " \
                f"reference: {self.reference}"

            if not self.is_saved_on(destination_realm):
                raise RuntimeError(
                    f"write_out({self}): modified object got lost "
                    f"on {destination_realm}")

            logger.debug(
                f"write_out({self}): not needed, object already"
                f" saved on {destination_realm}")
            return self.reference

        if self.realm:
            destination_realm = destination_realm or self.realm

        assert destination_realm is not None, \
            f"write_out({self}): no destination available for {self}"

        if Reference.is_remote(destination_realm):
            raise RuntimeError(
                f"write_out({self}): trying to write to a remote realm: "
                f"{destination_realm}")

        self.realm = destination_realm

        if self.is_saved_on(destination_realm):
            if not force_write:
                logger.debug(
                    f"write_out({self}): skipping map_out because {self} "
                    f"is already stored on {destination_realm}")
                return self.reference

            logger.debug(
                f"write_out({self}): forcing map_out, although {self} "
                f"is already stored on {destination_realm}")

        logger.debug(
            f"write_out({self}): calling map_out to save {self} "
            f"to {destination_realm}")

        self.reference = get_mapper(
            type(self).__name__,
            backend_type).map_out(
                self,
                destination_realm,
                force_write)

        self.set_saved_on(destination_realm)

        assert isinstance(self.reference, Reference), \
            f"write_out({self}): object {self} has no valid " \
            f"reference: {self.reference}"

        return self.reference

    def purge(self):
        if self.mapped:

            if len(self.saved_on) == 0:
                raise ValueError(
                    f"purge({self}): called with unsaved object: {self}")

            self.purge_impl()
            self.mapped = False

    def ensure_mapped(self,
                      backend_type="git") -> bool:
        if not self.mapped:
            self.read_in(backend_type)
            self.mapped = True
            return True
        return False

    def deepcopy(self,
                 new_mapper_family: Optional[str] = None,
                 new_destination: Optional[str] = None,
                 **kwargs) -> "MappableObject":

        with ensure_mapped(self):
            result = self.deepcopy_impl(new_mapper_family,
                                        new_destination,
                                        **kwargs)
        return result

    @abstractmethod
    def modifiable_sub_objects_impl(self) -> Iterable["MappableObject"]:
        raise NotImplementedError

    @abstractmethod
    def deepcopy_impl(self,
                      new_mapper_family: Optional[str] = None,
                      new_destination: Optional[str] = None,
                      **kwargs) -> "MappableObject":
        raise NotImplementedError

    @abstractmethod
    def purge_impl(self):
        raise NotImplementedError



import shlex
import subprocess
from typing import Any, List, Optional, Tuple, Union


def execute(arguments: Union[str, List[str]],
            stdin_content: Optional[Union[str, bytes]] = None) -> Any:

    return subprocess.run(
        shlex.split(arguments) if isinstance(arguments, str) else arguments,
        input=stdin_content.encode() if isinstance(stdin_content, str) else stdin_content,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)


def checked_execute(arguments: Union[str, List[str]],
                    stdin_content: Optional[Union[str, bytes]] = None) -> Tuple[str, str]:

    result = execute(arguments, stdin_content)
    if result.returncode != 0:
        raise RuntimeError(
            f"Command failed (exit code: {result.returncode}) {' '.join(arguments)}:\n"
            f"STDOUT:\n"
            f"{result.stdout.decode()}"
            f"STDERR:\n"
            f"{result.stderr.decode()}")
    return result.stdout.decode(), result.stderr.decode()



from uuid import UUID

from .gitbackend.subprocess import (
    git_ls_tree,
    git_save_tree,
    git_update_ref
)
from .objectreference import GitReference
from .utils import split_git_lstree_line
from ..mapper import Mapper
from ..reference import Reference


class UUIDSetGitMapper(Mapper):

    def map_in_impl(self,
                    uuid_set: "UUIDSet",
                    realm: str,
                    reference: Reference) -> None:

        from dataladmetadatamodel.uuidset import UUIDSet
        from dataladmetadatamodel.versionlist import VersionList

        assert isinstance(uuid_set, UUIDSet)
        assert isinstance(realm, str)
        assert isinstance(reference, Reference)

        uuid_set.uuid_set = dict()
        for line in git_ls_tree(realm, reference.location):
            line_elements = split_git_lstree_line(line)
            version_list = VersionList(
                realm=realm,
                reference=Reference("VersionList", line_elements[2]))

            uuid_set.uuid_set[UUID(line_elements[3])] = version_list

    def map_out_impl(self,
                     uuid_set: "UUIDSet",
                     realm: str,
                     force_write: bool) -> Reference:

        from dataladmetadatamodel.uuidset import UUIDSet
        assert isinstance(uuid_set, UUIDSet)

        tree_entries = [
            (
                "100644",
                "blob",
                version_list.write_out(realm).location,
                str(uuid)
            )
            for uuid, version_list in uuid_set.uuid_set.items()
        ]
        location = git_save_tree(realm, set(tree_entries))
        git_update_ref(realm, GitReference.UUID_SET.value, location)
        return Reference("UUIDSet", location)



"""
Commonly used functionality
"""
import time
from pathlib import Path
from typing import (
    Optional,
    Tuple,
    Union,
)
from uuid import UUID

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.metadata import Metadata
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.uuidset import UUIDSet
from dataladmetadatamodel.versionlist import (
    TreeVersionList,
    VersionList,
)
from dataladmetadatamodel.mapper.reference import Reference


def _get_tree_version_list_reference(location: str
                                     ) -> Reference:

    return Reference("TreeVersionList", location)


def _get_uuid_set_reference(
        mapper_family: str,
        realm: str,
        location: str) -> Reference:

    return Reference(
        "UUIDSet",
        location)


def get_top_level_metadata_objects(
        mapper_family: str,
        realm: Union[str, Path]
        ) -> Tuple[Optional[TreeVersionList], Optional[UUIDSet]]:

    """
    Load the two top-level metadata elements from a realm, i.e.
    the tree version list and the uuid list.

    We do this be creating references from known locations
    in the mapper family and loading the referenced objects.
    """
    from dataladmetadatamodel.mapper import (
        get_uuid_set_location,
        get_tree_version_list_location)

    realm = str(realm)

    tree_version_list = TreeVersionList(
        realm=realm,
        reference=_get_tree_version_list_reference(
            get_tree_version_list_location(mapper_family)))

    uuid_set = UUIDSet(
        realm=realm,
        reference=_get_uuid_set_reference(
            mapper_family,
            realm,
            get_uuid_set_location(mapper_family)))

    try:
        return tree_version_list.read_in(), uuid_set.read_in()
    except RuntimeError:
        return None, None


def get_top_nodes_and_metadata_root_record(
        mapper_family: str,
        realm: str,
        dataset_id: UUID,
        primary_data_version: str,
        prefix_path: MetadataPath,
        dataset_tree_path: MetadataPath,
        sub_dataset_id: Optional[UUID],
        sub_dataset_version: Optional[str],
        auto_create: Optional[bool] = False
) -> Tuple[Optional[TreeVersionList], Optional[UUIDSet], Optional[MetadataRootRecord]]:

    """
    Return the top nodes and metadata root record for a given dataset id
    and a given dataset version.

    If auto_create is True, create missing entries for the given dataset
    version, down to and including the metadata root record with empty
    dataset metadata and an empty file-tree connected to it.

    If auto_create is False and any of the elements: uuid set, version tree
    list, or metadata root record does not exist, (None, None, None) is
    returned.

    If you want to preserve any changes, are created metadata objects,
    you have to save the returned objects and flush the realm
    """
    tree_version_list, uuid_set = get_top_level_metadata_objects(
        mapper_family,
        realm)

    if tree_version_list is None and auto_create:
        tree_version_list = TreeVersionList()

    if uuid_set is None and auto_create:
        uuid_set = UUIDSet()

    if uuid_set is None or tree_version_list is None:
        return None, None, None

    metadata_root_record = get_metadata_root_record_from_top_nodes(
        tree_version_list,
        uuid_set,
        dataset_id,
        primary_data_version,
        prefix_path,
        dataset_tree_path,
        sub_dataset_id,
        sub_dataset_version,
        auto_create)

    return tree_version_list, uuid_set, metadata_root_record


def get_metadata_root_record_from_top_nodes(
        tree_version_list: TreeVersionList,
        uuid_set: UUIDSet,
        dataset_id: UUID,
        primary_data_version: str,
        prefix_path: MetadataPath,
        dataset_tree_path: MetadataPath,
        sub_dataset_id: Optional[UUID],
        sub_dataset_version: Optional[str],
        auto_create: Optional[bool] = False
) -> Optional[MetadataRootRecord]:

    """
    The metadata root record for a given dataset id and a given dataset version.

    If auto_create is True, create the metadata root record with empty
    dataset metadata and an empty file-tree connected to it.

    If auto_create is False and the metadata root record does not exist, None
    is returned.

    If you want to preserve any changes, to the created metadata objects,
    you have to save the returned objects (or their containers) and flush the
    realm.
    """
    if uuid_set is None or tree_version_list is None:
        return None

    if dataset_id in uuid_set.uuids():
        uuid_version_list = uuid_set.get_version_list(dataset_id)
    else:
        if auto_create is False:
            return None
        uuid_version_list = VersionList()
        uuid_set.set_version_list(dataset_id, uuid_version_list)

    # Get the dataset tree
    if (primary_data_version, prefix_path) in tree_version_list.versions_and_prefix_paths():
        time_stamp, prefix_path, dataset_tree = \
            tree_version_list.get_dataset_tree(primary_data_version, prefix_path)
    else:
        if auto_create is False:
            return None
        time_stamp = str(time.time())
        dataset_tree = DatasetTree()
        # TODO: set prefix_path
        # TODO: distinguish between unversioned prefix_path and dataset tree prefix_path
        tree_version_list.set_dataset_tree(
            primary_data_version=primary_data_version,
            time_stamp=time_stamp,
            prefix_path=prefix_path,
            dataset_tree=dataset_tree)

    if dataset_tree_path not in dataset_tree:
        if auto_create is False:
            return None

        dataset_level_metadata = Metadata()
        file_tree = FileTree()
        if dataset_tree_path != MetadataPath(""):
            assert sub_dataset_id is not None
            assert sub_dataset_version is not None
            metadata_root_record = MetadataRootRecord(
                sub_dataset_id,
                sub_dataset_version,
                dataset_level_metadata,
                file_tree)
        else:
            metadata_root_record = MetadataRootRecord(
                dataset_id,
                primary_data_version,
                dataset_level_metadata,
                file_tree)
        dataset_tree.add_dataset(dataset_tree_path, metadata_root_record)
    else:
        metadata_root_record = dataset_tree.get_metadata_root_record(
            dataset_tree_path)

    uuid_version_list.set_versioned_element(
        primary_data_version,
        str(time.time()),
        dataset_tree_path,
        metadata_root_record)

    return metadata_root_record



import subprocess
import tempfile
import unittest
from unittest.mock import patch

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.mapper.gitmapper.gitblobcache import hash_blob
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.tests.utils import (
    get_location,
    get_uuid,
    assert_dataset_trees_equal,
    create_dataset_tree
)


file_test_paths = [
    MetadataPath("a/b/c"),
    MetadataPath("a/b/a"),
    MetadataPath("b"),
    MetadataPath("c/d/e"),
    MetadataPath("a/x")]

dataset_test_paths = [
    MetadataPath("d1"),
    MetadataPath("d1/d1.1"),
    MetadataPath("d2"),
    MetadataPath("d2/d2.1/d2.1.1"),
    MetadataPath("d3/d3.1")]


uuid_0 = get_uuid(0)


class TestDatasetTree(unittest.TestCase):
    def test_add_metadata(self):

        paths = [
            MetadataPath("a/b/c"),
            MetadataPath("a/b/a"),
            MetadataPath("b"),
            MetadataPath("c/d/e")]

        mrr = MetadataRootRecord(uuid_0, "00112233", None, None)
        dataset_tree = create_dataset_tree(
            dataset_tree_paths=paths,
            initial_mrr=mrr)

        returned_entries = tuple(dataset_tree.dataset_paths)

        returned_paths = [entry[0] for entry in returned_entries]
        self.assertEqual(sorted(paths), sorted(returned_paths))

        for entry in returned_entries:
            self.assertEqual(entry[1], mrr)

    def test_root_node(self):

        dataset_tree = DatasetTree()

        mrr = MetadataRootRecord(uuid_0, "00112233", None, None)
        dataset_tree.add_dataset(MetadataPath(""), mrr)
        self.assertEqual(dataset_tree.get_metadata_root_record(MetadataPath("")), mrr)

        returned_entries = tuple(dataset_tree.dataset_paths)
        self.assertEqual(len(returned_entries), 1)

        self.assertEqual(returned_entries[0][0], MetadataPath(""))
        self.assertEqual(returned_entries[0][1], mrr)


class TestReferenceCreation(unittest.TestCase):
    def test_object_reference_creation(self):

        dataset_tree = create_dataset_tree(dataset_test_paths)
        with \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "metadatamapper.git_save_str") as save_str, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "mtreenodemapper.git_save_tree_node") as save_tree_node, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "metadatarootrecordmapper.git_save_json") as save_json, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "gitblobcache.git_save_file_list") as file_list_save, \
                patch("dataladmetadatamodel.mtreeproxy."
                      "add_tree_reference") as add_tree_ref:

            save_str.return_value = get_location(1)
            save_tree_node.return_value = get_location(2)
            save_json.return_value = get_location(3)
            file_list_save.side_effect = lambda r, l: [
                hash_blob(open(e, "rb").read())
                for e in l
            ]

            dataset_tree.write_out("/tmp/t1")

            # We expect one call for the dataset-tree itself
            # and one call for each file-tree, one of which
            # is anchored at each dataset prefix_path
            self.assertEqual(
                add_tree_ref.call_count,
                1 + len(dataset_test_paths)
            )


class TestDeepCopy(unittest.TestCase):

    def test_copy_from_memory(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            dataset_tree = create_dataset_tree(dataset_test_paths,
                                               file_test_paths)

            dataset_tree_copy = dataset_tree.deepcopy("git", copy_dir)
            dataset_tree_copy.read_in()

            assert_dataset_trees_equal(
                self,
                dataset_tree,
                dataset_tree_copy,
                True)

    def test_copy_from_backend(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            dataset_tree = create_dataset_tree(dataset_test_paths,
                                               file_test_paths)

            dataset_tree.write_out(original_dir)

            dataset_tree_copy = dataset_tree.deepcopy("git", copy_dir)
            dataset_tree_copy.read_in()

            assert_dataset_trees_equal(
                self,
                dataset_tree,
                dataset_tree_copy,
                True)


class TestSubTreeManipulation(unittest.TestCase):

    def get_mrr(self, n: int = 0) -> MetadataRootRecord:
        return MetadataRootRecord(uuid_0, f"00112233-{n}", None, None)

    def test_subtree_adding(self):
        mrr_1 = self.get_mrr(1)
        mrr_2 = self.get_mrr(2)

        tree = DatasetTree()
        tree.add_dataset(MetadataPath("a/b/c"), mrr_1)

        subtree = DatasetTree()
        subtree.add_dataset(MetadataPath("d/e/f"), mrr_2)

        tree.add_subtree(subtree, MetadataPath("a/x"))

        mrr = tree.get_metadata_root_record(MetadataPath("a/b/c"))
        self.assertEqual(mrr, mrr_1)

        mrr = tree.get_metadata_root_record(MetadataPath("a/x/d/e/f"))
        self.assertEqual(mrr, mrr_2)

    def test_subtree_adding_with_conversion(self):
        mrr_1 = self.get_mrr()
        mrr_2 = self.get_mrr()

        tree = DatasetTree()
        tree.add_dataset(MetadataPath("a/b/c"), mrr_1)

        subtree = DatasetTree()
        subtree.add_dataset(MetadataPath("e/f"), mrr_2)

        tree.add_subtree(subtree, MetadataPath("a/b/c/d"))

        node = tree.get_metadata_root_record(MetadataPath("a/b/c"))
        self.assertEqual(node, mrr_1)

        node = tree.get_metadata_root_record(MetadataPath("a/b/c/d/e/f"))
        self.assertEqual(node, mrr_2)

    def test_subtree_adding_on_existing_path(self):
        tree = DatasetTree()
        tree.add_dataset(MetadataPath("a/b/c/d"), self.get_mrr())

        subtree = DatasetTree()
        subtree.add_dataset(MetadataPath("e/f"), self.get_mrr())

        self.assertRaises(
            ValueError,
            tree.add_subtree,
            subtree, MetadataPath("a/b/c/d"))

    def test_subtree_deletion(self):
        mrr_1 = self.get_mrr()
        mrr_2 = self.get_mrr()

        tree = DatasetTree()
        tree.add_dataset(MetadataPath("a/b/c"), mrr_1)
        tree.add_dataset(MetadataPath("a/b/c/d/e/f"), mrr_2)

        self.assertIsNotNone(tree.get_metadata_root_record(MetadataPath("a/b/c/d/e/f")))
        tree.delete_subtree(MetadataPath("a/b/c/d/e/f"))
        self.assertIsNone(tree.get_metadata_root_record(MetadataPath("a/b/c/d/e/f")))
        self.assertIsNotNone(tree.mtree.get_object_at_path(MetadataPath("a/b/c/d/e")))

        tree.delete_subtree(MetadataPath("a/b/c"))
        self.assertIsNotNone(tree.mtree.get_object_at_path(MetadataPath("a/b")))
        self.assertIsNone(tree.get_metadata_root_record(MetadataPath("a/b/c")))
        self.assertIsNone(tree.get_metadata_root_record(MetadataPath("a/b/c/d")))
        self.assertIsNone(tree.get_metadata_root_record(MetadataPath("a/b/c/d/e")))


if __name__ == '__main__':
    unittest.main()



import codecs
import subprocess
import tempfile
import unittest
from typing import List
from unittest import mock

from ..mtreenodemapper import MTreeNodeGitMapper
from ...reference import Reference
from ....mtreenode import MTreeNode
from ....text import Text


tree_location = "a000000000000000000000000000000000000000"
file_location = "a001000000000000000000000000000000000001"


default_file_names = [
    "a a",
    "b b",
    ";< |F|| >- \\.",
    "🐶🐷",
    "\\F\\",
    "\\ .dat\\|",
    '"',
    '\\"',
    "\\",
    "'",
]

default_sub_dir_names = [
    "sub 0",
    "sub 1",
    "sub 2",
    ";< [A] ||| >- \\.",
    "🐶 [B] 🐷 .datc",
    "\\[C]",
]


def create_tree(file_names: List[str],
                sub_dir_names: List[str]
                ) -> MTreeNode:

    root_node = MTreeNode(Text)
    for sub_dir_name in sub_dir_names:
        sub_tree_node = MTreeNode(Text)
        for file_name in file_names:
            sub_tree_node.add_child(
                file_name,
                Text(f"content of: /{sub_dir_name}/{file_name}"))
        root_node.add_child(sub_dir_name, sub_tree_node)
    return root_node


class TestMTreeNodeMapper(unittest.TestCase):

    def test_basic_unmapping(self):

        realm = "/tmp/t1"
        file_names = default_file_names
        sub_dir_names = default_sub_dir_names

        root_node = create_tree(file_names, sub_dir_names)

        with \
                mock.patch(
                    "dataladmetadatamodel.mapper.gitmapper"
                    ".mtreenodemapper.git_save_tree_node") as save_tree_node, \
                mock.patch(
                    "dataladmetadatamodel.mapper.gitmapper" 
                    ".textmapper.git_save_str") as save_str:

            save_tree_node.configure_mock(return_value=tree_location)
            save_str.configure_mock(return_value=file_location)

            root_node.write_out(realm, "git")

            # Expect one (root) plus len(sub-dir trees) calls
            assert save_tree_node.call_count == 1 + len(sub_dir_names)

            expected_sub_tree_calls = [
                mock.call(
                    realm,
                    [
                        (
                            "100644",
                            "blob",
                            file_location,
                            MTreeNodeGitMapper.escape_double_quotes(file_name)
                        )
                        for file_name in file_names
                    ]
                )
                for _ in range(len(sub_dir_names))
            ]
            save_tree_node.assert_has_calls(
                expected_sub_tree_calls,
                any_order=True
            )

            expected_root_calls = [
                mock.call(
                    realm,
                    [
                        (
                            "040000",
                            "tree",
                            tree_location,
                            MTreeNodeGitMapper.escape_double_quotes(sub_dir_name)
                        )
                        for sub_dir_name in sub_dir_names
                    ]
                )
            ]
            save_tree_node.assert_has_calls(expected_root_calls, any_order=True)

    def test_none_mapping_out(self):
        tree = MTreeNode(Text)
        reference = tree.write_out("/tmp/t1")
        self.assertTrue(reference.is_none_reference())

    def test_none_mapping_in(self):
        reference = Reference.get_none_reference("MTreeNode")
        tree = MTreeNode(Text, "/tmp/t1", reference)
        tree.read_in()
        self.assertEqual(len(tree.child_nodes), 0)

    def test_mapping_end_to_end(self):
        file_names = default_file_names
        sub_dir_names = default_sub_dir_names

        with tempfile.TemporaryDirectory() as realm:

            subprocess.run(["git", "init", realm])

            root_node = create_tree(file_names, sub_dir_names)
            reference = root_node.write_out(realm)

            new_root_node = MTreeNode(
                leaf_class=Text,
                realm=realm,
                reference=reference)
            new_root_node.read_in()

            tree_elements = list(new_root_node.get_paths_recursive())
            assert len(tree_elements) == len(file_names) * len(sub_dir_names)

            assert (
                sorted([str(tree_element[0]) for tree_element in tree_elements])
                == sorted([
                    sub_dir_name + "/" + file_name
                    for sub_dir_name in sub_dir_names
                    for file_name in file_names
                ])
            )

            assert (
                sorted([
                    tree_element[1].read_in().content
                    for tree_element in tree_elements
                ])
                == sorted([
                    f"content of: /{sub_dir_name}/{file_name}"
                    for sub_dir_name in sub_dir_names
                    for file_name in file_names
                ])
            )


def test_decode_assumptions():
    encoded_decoded_pairs = [
        ('\\"', '"'),
        ('\\\\', '\\'),
        ('\\303\\204', 'Ä'),
    ]
    for encoded, decoded in encoded_decoded_pairs:
        assert codecs.escape_decode(encoded)[0].decode("utf-8") == decoded


if __name__ == '__main__':
    unittest.main()



import subprocess
import tempfile
import time
import unittest
from pathlib import Path
from typing import cast
from unittest.mock import patch

from dataladmetadatamodel.filetree import FileTree
from dataladmetadatamodel.metadata import (
    ExtractorConfiguration,
    Metadata,
    MetadataInstance,
)
from dataladmetadatamodel.mapper.gitmapper.gitblobcache import hash_blob
from dataladmetadatamodel.mapper.gitmapper.objectreference import flush_object_references
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.tests.utils import (
    assert_file_trees_equal,
    create_file_tree_with_metadata,
    get_location,
)


default_paths = [
    MetadataPath("/a/b/c"),
    MetadataPath("a/b/a"),
    MetadataPath("b"),
    MetadataPath("c/d/e"),
    MetadataPath("a/x"),
]


class TestFileTree(unittest.TestCase):

    def test_add_metadata(self):

        file_tree = create_file_tree_with_metadata(default_paths, [
            Metadata()
            for _ in default_paths])

        returned_entries = tuple(file_tree.get_paths_recursive())

        returned_paths = [entry[0] for entry in returned_entries]
        self.assertEqual(sorted(default_paths), sorted(returned_paths))

        for returned_path, returned_metadata in [
                (entry[0], entry[1])
                for entry in returned_entries]:
            self.assertEqual(returned_metadata, Metadata())

    def test_add_extractor_run(self):

        file_tree = create_file_tree_with_metadata(default_paths, [
            Metadata()
            for _ in default_paths])

        author_name = "Karl-Test"
        author_email = author_name + "@test.com"
        extractor_name = "test_extractor"
        extractor_configuration = ExtractorConfiguration(
            "extractor_version_1",
            {"key1": "value1"})
        metadata_content = {"key0": "this is metadata"}

        file_tree.add_extractor_run(
            default_paths[0],
            1.2,
            extractor_name,
            author_name,
            author_email,
            extractor_configuration,
            metadata_content)

        metadata = file_tree.get_metadata(default_paths[0])
        self.assertIsNotNone(metadata)

        stored_metadata = metadata.extractor_runs_for_extractor(extractor_name)

        self.assertEqual(
            stored_metadata.parameter_set,
            [extractor_configuration])

        self.assertEqual(
            stored_metadata._instances[0],
            MetadataInstance(
                1.2,
                author_name,
                author_email,
                extractor_configuration,
                metadata_content))


class TestReferenceCreation(unittest.TestCase):
    def test_object_reference_creation(self):

        file_tree = create_file_tree_with_metadata(
            default_paths,
            [
                Metadata()
                for _ in default_paths
            ]
        )

        with \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "metadatamapper.git_save_str") as save_str, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "mtreenodemapper.git_save_tree_node") as save_tree_node, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "metadatarootrecordmapper.git_save_json") as save_json, \
                patch("dataladmetadatamodel.mapper.gitmapper."
                      "gitblobcache.git_save_file_list") as file_list_save, \
                patch("dataladmetadatamodel.mtreeproxy."
                      "add_tree_reference") as add_tree_ref:

            save_str.return_value = get_location(1)
            save_tree_node.return_value = get_location(2)
            save_json.return_value = get_location(3)
            file_list_save.side_effect = lambda r, l: [
                hash_blob(open(e, "rb").read())
                for e in l
            ]

            file_tree.write_out("/tmp/t1")

            # We expect one call for the dataset-tree itself
            # and one call for each file-tree, one of which
            # is anchored at each dataset prefix_path
            add_tree_ref.assert_called_once()


class TestMapping(unittest.TestCase):

    def test_adding(self):
        # check file tree adding is working
        with tempfile.TemporaryDirectory() as metadata_store:

            subprocess.run(["git", "init", metadata_store])

            file_tree = create_file_tree_with_metadata(
                default_paths,
                [Metadata() for _ in default_paths])
            reference = file_tree.write_out(metadata_store)

            file_tree = cast(FileTree, FileTree(
                realm=metadata_store,
                reference=reference).read_in())

            additional_paths = [MetadataPath(f"x/y.{n}") for n in range(10)]
            for additional_path in additional_paths:
                file_tree.add_metadata(additional_path, Metadata())
            reference = file_tree.write_out()

            file_tree = FileTree(realm=metadata_store, reference=reference).read_in()
            read_paths = [pair[0] for pair in file_tree.get_paths_recursive()]
            for path in default_paths + additional_paths:
                self.assertIn(path, read_paths)

    def test_adding_to_massive_tree(self):
        # check file tree adding is working
        with tempfile.TemporaryDirectory() as metadata_store:

            subprocess.run(["git", "init", metadata_store])

            file_tree = FileTree()

            start_time = time.time()
            for first_part in range(10):
                for second_part in range(10):
                    for third_part in range(10):
                        metadata_path = MetadataPath(f"{first_part:03}/"
                                                     f"{second_part:03}/"
                                                     f"{third_part:03}")
                        file_tree.add_metadata(metadata_path, Metadata())
            initialisation_duration = time.time() - start_time
            print(f"Initialised: {initialisation_duration:4f}")

            start_time = time.time()
            reference = file_tree.write_out(metadata_store)
            write_out_duration = time.time() - start_time
            print(f"Written out: {write_out_duration:4f}")

            start_time = time.time()
            file_tree = FileTree(realm=metadata_store, reference=reference).read_in()
            read_in_duration = time.time() - start_time
            print(f"Read in: {read_in_duration:4f}")

            start_time = time.time()
            file_tree.add_metadata(MetadataPath("5/5/xxx"), Metadata())
            add_duration = time.time() - start_time
            print(f"Added single entry: {add_duration:4f}")

            start_time = time.time()
            file_tree.write_out()
            write_out_2nd_duration = time.time() - start_time
            print(f"Written out single entry: {write_out_2nd_duration:4f}")

    def test_shallow_file_tree_mapping(self):
        # assert that file trees content is not mapped by default
        with tempfile.TemporaryDirectory() as metadata_store:

            subprocess.run(["git", "init", metadata_store])

            paths = [
                MetadataPath("a"),
                MetadataPath("b")]

            file_tree = FileTree()
            for path in paths:
                metadata = Metadata()
                file_tree.add_metadata(path, metadata)
                file_tree.unget_metadata(metadata, metadata_store)

            reference = file_tree.write_out(metadata_store)
            flush_object_references(Path(metadata_store))

            new_file_tree = FileTree(realm=metadata_store, reference=reference).read_in()
            self.assertFalse(new_file_tree.mtree.child_nodes["a"].mapped)


class TestDeepCopy(unittest.TestCase):

    def test_copy_from_memory(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            file_tree = FileTree()
            for path in ["a/b/c/d", "a/b/d", "a/x"]:
                file_tree.add_metadata(
                    MetadataPath(path),
                    Metadata())

            file_tree_copy = file_tree.deepcopy(new_destination=copy_dir)
            assert_file_trees_equal(self, file_tree, file_tree_copy, True)

    def test_copy_from_backend(self):
        with \
                tempfile.TemporaryDirectory() as original_dir, \
                tempfile.TemporaryDirectory() as copy_dir:

            subprocess.run(["git", "init", original_dir])
            subprocess.run(["git", "init", copy_dir])

            paths = [
                MetadataPath("a/b/c/d"),
                MetadataPath("a/b/d"),
                MetadataPath("a/x")]

            file_tree = FileTree()
            for path in paths:
                metadata = Metadata()
                file_tree.add_metadata(path, metadata)
                file_tree.unget_metadata(metadata, original_dir)
            file_tree.write_out(original_dir)

            file_tree_copy = file_tree.deepcopy(new_destination=copy_dir)
            file_tree_copy.read_in()

            assert_file_trees_equal(self, file_tree, file_tree_copy, True)


if __name__ == '__main__':
    unittest.main()



import sys
import time
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.uuidset import UUIDSet
from dataladmetadatamodel.versionlist import VersionList

from tools.metadata_creator.mrrcreator import create_metadata_root_record
from tools.metadata_creator.utils import get_dataset_id, get_dataset_version, read_datasets


def create_uuid_set(realm: str,
                    path: str,
                    parameter_set_count: int
                    ) -> Optional[UUIDSet]:

    uuid_set = UUIDSet(realm=realm)
    for path, dir_entry in read_datasets(path):

        dataset_id = get_dataset_id(dir_entry.path)
        if dataset_id is None:
            print(f"cannot determine id of dataset at {dir_entry.path}", file=sys.stderr)
            continue

        version = get_dataset_version(dir_entry.path)
        if version is None:
            print(f"cannot determine version of dataset at {dir_entry.path}", file=sys.stderr)
            continue

        version_list = VersionList()
        version_list.set_versioned_element(
            version,
            str(time.time()),
            path,
            create_metadata_root_record(
                mapper_family,
                realm,
                dataset_id,
                version,
                dir_entry.path,
                path,
                parameter_set_count
            )
        )

        uuid_set.set_version_list(
            dataset_id,
            version_list
        )

    return uuid_set


def create_uuid_set_for_mrrs(mapper_family: str,
                             realm: str,
                             metadata_root_records: Dict[Tuple[UUID, str, str], MetadataRootRecord]
                             ) -> UUIDSet:

    uuid_version_list = dict()
    uuid_set = UUIDSet(mapper_family, realm)

    for id_version_path, metadata_root_record in metadata_root_records.items():

        uuid, dataset_version, relative_path = id_version_path

        version_list = uuid_version_list.get(
            uuid,
            VersionList())

        version_list.set_versioned_element(
            dataset_version,
            str(time.time()),
            relative_path,
            metadata_root_record
        )
        uuid_version_list[uuid] = version_list

    for uuid, version_list in uuid_version_list.items():
        uuid_set.set_version_list(uuid, version_list)

    return uuid_set



import subprocess
import tempfile
import unittest

from dataladmetadatamodel.common import (
    get_top_level_metadata_objects,
    get_top_nodes_and_metadata_root_record,
)
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.tests.utils import get_uuid
from dataladmetadatamodel.uuidset import UUIDSet
from dataladmetadatamodel.versionlist import TreeVersionList


uuid_0 = get_uuid(0)
uuid_1 = get_uuid(1)


class TestTopLevelObjects(unittest.TestCase):

    # This is not exactly a unit test since it does not
    # isolate the units from the other components, but
    # more a module test.
    def test_top_level_functions(self):

        with tempfile.TemporaryDirectory() as realm:
            subprocess.run(["git", "init", realm])

            tvl, uuid_set = get_top_level_metadata_objects("git", realm)
            self.assertIsNone(tvl)
            self.assertIsNone(uuid_set)

            tvl, uuid_set, mrr = get_top_nodes_and_metadata_root_record(
                mapper_family="git",
                realm=realm,
                dataset_id=uuid_0,
                primary_data_version="v1",
                prefix_path=MetadataPath(""),
                dataset_tree_path=MetadataPath("a/b/c"),
                sub_dataset_id=None,
                sub_dataset_version=None,
                auto_create=False
            )
            self.assertIsNone(tvl)
            self.assertIsNone(uuid_set)
            self.assertIsNone(mrr)

            tvl, uuid_set, mrr = get_top_nodes_and_metadata_root_record(
                mapper_family="git",
                realm=realm,
                dataset_id=uuid_0,
                primary_data_version="v1",
                prefix_path=MetadataPath(""),
                dataset_tree_path=MetadataPath("a/b/c"),
                sub_dataset_id=uuid_1,
                sub_dataset_version="v-sub1",
                auto_create=True
            )
            self.assertIsInstance(tvl, TreeVersionList)
            self.assertIsInstance(uuid_set, UUIDSet)
            self.assertIsInstance(mrr, MetadataRootRecord)

            tvl.write_out(realm)
            uuid_set.write_out(realm)

            tvl, uuid_set = get_top_level_metadata_objects("git", realm)
            self.assertIsInstance(tvl, TreeVersionList)
            self.assertIsInstance(uuid_set, UUIDSet)


if __name__ == '__main__':
    unittest.main()



import json
import logging
from typing import Optional

from dataladmetadatamodel import (
    check_serialized_version,
    version_string
)


logger = logging.getLogger("datalad.metadatamodel.mapper")

none_class_name = "*None*"
none_location = "*None*"


class Reference:
    def __init__(self,
                 class_name: str,
                 location: Optional[str] = None):

        assert isinstance(location, str) or location is None, \
                f"location is not a string: {location}"

        self.class_name = class_name
        self.location = location

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return (
            f"Reference(class_name='{self.class_name}', "
            f"location={repr(self.location)})")

    def __eq__(self, other):
        return (
            self.class_name == other.class_name
            and self.location == other.location
        )

    def assign_from(self, other: "Reference"):
        self.class_name = other.class_name
        self.location = other.location

    def is_none_reference(self) -> bool:
        return self.location == none_location

    def to_json_str(self):
        return json.dumps(self.to_json_obj())

    def to_json_obj(self):
        return {
            "@": dict(
                type="Reference",
                version=version_string),
            **dict(
                class_name=self.class_name,
                location=self.location)
        }

    @classmethod
    def from_json_str(cls, json_str: str) -> "Reference":
        return cls.from_json_obj(json.loads(json_str))

    @classmethod
    def from_json_obj(cls, obj) -> "Reference":
        assert obj["@"]["type"] == "Reference"
        check_serialized_version(obj)
        if "mapper_family" in obj or "realm" in obj:
            logger.info(f"old reference object found: {obj}")
        return cls(
            obj["class_name"],
            obj["location"]
        )

    @classmethod
    def get_none_reference(cls, referred_class_name: str) -> "Reference":
        return cls(referred_class_name, none_location)

    @staticmethod
    def is_remote(realm: str):
        return any(
            map(
                lambda pattern: realm.startswith(pattern),
                [
                   "git@",
                   "git:",
                   "http:",
                   "https:",
                   "ssh:"]))

    @staticmethod
    def is_local(realm: str):
        return not Reference.is_remote(realm)



import subprocess

from ..utils import split_git_lstree_line
from ..gitbackend.subprocess import (
    git_ls_tree,
    git_ls_tree_recursive,
)


def test_split_git_lstree_line(tmp_path):
    d = tmp_path / "dir 1"
    d.mkdir(parents=True)
    f = d / "file 1"
    f.write_text("some content")

    subprocess.run(["git", "init"], cwd=str(tmp_path))
    subprocess.run(["git", "add", "dir 1"], cwd=str(tmp_path))
    subprocess.run(["git", "commit", "-m", "some stuff"], cwd=str(tmp_path))

    l = git_ls_tree(str(tmp_path), "HEAD")[0]
    _, _, _, name = split_git_lstree_line(l)
    assert name == "dir 1"

    l = git_ls_tree_recursive(str(tmp_path), "HEAD")[0]
    _, _, _, name = split_git_lstree_line(l)
    assert name == "dir 1/file 1"



import enum
import logging
from pathlib import Path
from typing import (
    Set,
    Tuple,
)

from ...mapper.reference import none_location
from .treeupdater import EntryType
from .utils import split_git_lstree_line


logger = logging.getLogger("datalad.metalad.gitmapper.objectreference")

cached_object_references: Set[Tuple[EntryType, str]] = set()


class GitReference(enum.Enum):
    TREE_VERSION_LIST = "refs/datalad/dataset-tree-version-list"
    UUID_SET = "refs/datalad/dataset-uuid-set"
    OBJECT_REFERENCES = "refs/datalad/object-references-2.0"
    LEGACY_TREES = "refs/datalad/object-references/trees"
    LEGACY_BLOBS = "refs/datalad/object-references/blobs"


checked_legacy_stores = set()


def add_object_reference(entry_type: EntryType,
                         object_hash: str):

    if object_hash == none_location:
        logger.warning("attempt to add a None-reference")
        return

    cached_object_references.add((entry_type, object_hash))


def add_legacy_store_entries_from(realm: Path,
                                  reference: GitReference,
                                  entry_type: EntryType):

    from .gitbackend.subprocess import git_ls_tree

    try:
        legacy_entries = (
            split_git_lstree_line(line)
            for line in git_ls_tree(str(realm), reference.value))

        logger.info(f"converting legacy reference store at {realm} (ref: "
                    f"{reference.value}) to new format")

        for entry in legacy_entries:
            cached_object_references.add((entry_type, entry[2]))
        return True
    except RuntimeError:
        return False


def add_legacy_store_entries(realm: Path) -> Tuple[bool, bool]:
    """
    Add legacy store entries to the cache, if they exist
    """
    if realm not in checked_legacy_stores:
        checked_legacy_stores.add(realm)

        legacy_trees_added = add_legacy_store_entries_from(
            realm,
            GitReference.LEGACY_TREES,
            EntryType.Directory)

        legacy_blobs_added = add_legacy_store_entries_from(
            realm,
            GitReference.LEGACY_BLOBS,
            EntryType.File)

        return legacy_trees_added, legacy_blobs_added

    return False, False


def remove_legacy_store_reference(realm: Path, reference: GitReference):
    from .gitbackend.subprocess import git_delete_ref

    logger.info(
        f"deleting legacy object reference store: {realm} "
        f"(ref: {reference.value})")
    git_delete_ref(str(realm), reference.value)


def flush_object_references(realm: Path):
    global cached_object_references

    from .gitbackend.subprocess import git_update_ref
    from ..gitmapper.treeupdater import (
        PathInfo,
        _get_dir,
        add_paths,
    )
    from ..gitmapper.utils import locked_backend

    legacy_trees_added, legacy_blobs_added = add_legacy_store_entries(realm)

    if cached_object_references:
        path_infos = [
            PathInfo(
                [
                    object_hash[0:3],
                    object_hash[3:6],
                    object_hash[6:9],
                    object_hash[9:],
                ],
                object_hash,
                entry_type)
            for entry_type, object_hash in cached_object_references
        ]

        with locked_backend(realm):
            try:
                root_entries = _get_dir(realm, GitReference.OBJECT_REFERENCES.value)
            except RuntimeError:
                root_entries = []

            tree_hash = add_paths(realm, path_infos, root_entries)
            git_update_ref(str(realm), GitReference.OBJECT_REFERENCES.value, tree_hash)

        cached_object_references = set()

    if legacy_trees_added is True:
        remove_legacy_store_reference(realm, GitReference.LEGACY_TREES)
    if legacy_blobs_added is True:
        remove_legacy_store_reference(realm, GitReference.LEGACY_BLOBS)


def add_tree_reference(object_hash: str):
    add_object_reference(EntryType.Directory, object_hash)


def add_blob_reference(object_hash: str):
    add_object_reference(EntryType.File, object_hash)


def remove_object_reference(*args, **kwargs):
    raise NotImplementedError



from typing import (
    Iterable,
    Optional,
    Tuple
)

from dataladmetadatamodel import JSONObject
from dataladmetadatamodel.metadata import (
    ExtractorConfiguration,
    Metadata
)
from dataladmetadatamodel.mappableobject import MappableObject
from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.mtreenode import MTreeNode
from dataladmetadatamodel.mtreeproxy import MTreeProxy
from dataladmetadatamodel.mapper.reference import Reference


class FileTree(MTreeProxy):
    def __init__(self,
                 mtree: Optional[MTreeNode] = None,
                 realm: Optional[str] = None,
                 reference: Optional[Reference] = None):

        assert isinstance(mtree, (type(None), MTreeNode))
        assert isinstance(reference, (type(None), Reference))

        super().__init__(Metadata, mtree, realm, reference)

    def add_metadata(self,
                     path: MetadataPath,
                     metadata: Metadata):

        self.mtree.add_child_at(metadata, path)

    def get_metadata(self,
                     path: MetadataPath
                     ) -> Optional[Metadata]:

        metadata = self.mtree.get_object_at_path(path)

        if metadata is None:
            return None

        assert isinstance(metadata, Metadata)
        metadata.ensure_mapped()
        return metadata

    def unget_metadata(self,
                       metadata: Metadata,
                       destination: Optional[str] = None):
        assert isinstance(metadata, Metadata)
        metadata.write_out(destination)
        metadata.purge()

    def add_extractor_run(self,
                          path,
                          time_stamp: Optional[float],
                          extractor_name: str,
                          author_name: str,
                          author_email: str,
                          configuration: ExtractorConfiguration,
                          metadata_content: JSONObject):

        try:
            metadata = self.get_metadata(path)
        except AttributeError:
            metadata = Metadata()
            self.add_metadata(path, metadata)

        metadata.add_extractor_run(
            time_stamp,
            extractor_name,
            author_name,
            author_email,
            configuration,
            metadata_content
        )

    def get_paths_recursive(self,
                            show_intermediate: Optional[bool] = False
                            ) -> Iterable[Tuple[MetadataPath, MappableObject]]:

        yield from self.mtree.get_paths_recursive(show_intermediate)



import subprocess
import tempfile
from hashlib import sha1
from pathlib import Path
from typing import (
    List,
    Iterable,
    Tuple,
)

from ..gitbackend.subprocess import (
    git_ls_tree,
    git_save_tree_node,
    git_update_ref,
)
from ..objectreference import (
    GitReference,
    flush_object_references,
)
from ..treeupdater import EntryType
from ..utils import (
    create_git_repo,
    split_git_lstree_line,
)


def _does_ref_exist(realm: Path, reference: str) -> bool:
    result = subprocess.run(
        ["git", "--git-dir", str(realm / ".git"), "show-ref", reference])
    return result.returncode == 0


def _create_hash(n: int) -> str:
    return sha1(f"index: {n}".encode()).hexdigest()


def _create_legacy_tree(start_index: int,
                        stop_index: int,
                        entry_type: EntryType
                        ) -> Iterable[Tuple[str, str, str, str]]:

    flag, object_type = {
        EntryType.File: ("100644", "blob"),
        EntryType.Directory: ("040000", "tree")
    }[entry_type]

    return (
        (
            flag,
            object_type,
            _create_hash(i),
            "object_reference:" + _create_hash(i)
         )
        for i in range(start_index, stop_index)
    )


def _write_legacy_tree(realm: Path,
                       reference: str,
                       entry_type: EntryType,
                       start_index: int,
                       stop_index: int,
                       ):

    tree_hash = git_save_tree_node(
        str(realm),
        _create_legacy_tree(start_index, stop_index, entry_type))
    git_update_ref(str(realm), reference, tree_hash)


def _read_tree(realm: Path, reference: str) -> Iterable[Tuple[str, str]]:
    return (
        (element[3], element[2])
        for element in (
            split_git_lstree_line(line)
            for line in git_ls_tree(str(realm), reference)))


def _read_reference_tree(realm: Path,
                         reference: str,
                         parts: List[str]
                         ) -> List[str]:
    """A simple recursive reference tree reader"""

    if len(parts) == 4:
        assert reference[9:] == parts[3]
        return ["".join(parts)]

    result = []
    for name, sub_reference in _read_tree(realm, reference):
        sub_results = _read_reference_tree(realm, sub_reference, parts + [name])
        result.extend(sub_results)
    return result


def test_old_reference_conversion():
    with tempfile.TemporaryDirectory() as td:
        realm = Path(td)
        create_git_repo(realm, {"readme.md": "test repo"})

        _write_legacy_tree(realm, GitReference.LEGACY_BLOBS.value, EntryType.File, 100, 200)
        _write_legacy_tree(realm, GitReference.LEGACY_TREES.value, EntryType.Directory, 200, 300)

        flush_object_references(realm)

        references = _read_reference_tree(
            realm,
            GitReference.OBJECT_REFERENCES.value,
            [])

        # Check that the new references exist.
        for i in range(100, 300):
            assert _create_hash(i) in references

        # Check that the old references are gone by trying to update them
        assert _does_ref_exist(realm, GitReference.LEGACY_BLOBS.value) is False
        assert _does_ref_exist(realm, GitReference.LEGACY_TREES.value) is False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
import sys
from os.path import dirname

# This is needed for versioneer to be importable when building with PEP 517.
# See <https://github.com/warner/python-versioneer/issues/193> and links
# therein for more information.
sys.path.append(dirname(__file__))
import versioneer


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="datalad-metadata-model",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="The Datalad Team",
    author_email="christian.moench@web.de",
    description="Datalad Metadata Model",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/christian-monch/metadata-model",
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "mdc=tools.metadata_creator.main:main"
        ]
    }
)



import subprocess
import tempfile
import time
import unittest

from dataladmetadatamodel.metadatapath import MetadataPath
from dataladmetadatamodel.mtreenode import MTreeNode
from dataladmetadatamodel.text import Text


default_paths = [
    "x",
    "a/0",
    "a/1",
    "a/b/0",
    "a/b/1",
    "a/b/2",
    "a a/b b b/2 2 2 2",
    ";< ||| >- \\./ 🐶🐷/\\/\\ .datc"
]


class TestMTreeNode(unittest.TestCase):

    def test_basic(self):
        MTreeNode(Text)

    def test_adding(self):
        mtree_node = MTreeNode(Text)

        for path in default_paths:
            mtree_node.add_child_at(Text(f"content of: {path}"),
                                    MetadataPath(path))

        all_children = list(mtree_node.get_paths_recursive())
        self.assertEqual(len(default_paths), len(all_children))
        for path, child in all_children:
            self.assertIn(str(path), default_paths)
            self.assertIsInstance(child, Text)
            self.assertEqual(child.content, f"content of: {str(path)}")

    def test_get_paths_recursive(self):
        mtree_node = MTreeNode(Text)

        for path in default_paths:
            mtree_node.add_child_at(Text(f"content of: {path}"),
                                    MetadataPath(path))
        results = list(mtree_node.get_paths_recursive())
        returned_paths = set(map(lambda e: str(e[0]), results))
        self.assertSetEqual(set(default_paths), returned_paths)

    def test_get_paths_recursive_intermediate(self):
        mtree_node = MTreeNode(Text)

        for path in default_paths:
            mtree_node.add_child_at(Text(f"content of: {path}"),
                                    MetadataPath(path))
        results = list(mtree_node.get_paths_recursive(True))

        intermediate = set([
            "/".join(path.split("/")[:i])
            for path in default_paths
            for i in range(len(path))
        ] + default_paths)

        returned_paths = set(map(lambda e: str(e[0]), results))
        self.assertSetEqual(intermediate, returned_paths)


class TestMTreeNodeMapping(unittest.TestCase):

    def test_adding_to_massive_tree(self):
        with tempfile.TemporaryDirectory() as metadata_store:

            subprocess.run(["git", "init", metadata_store])

            mtree = MTreeNode(leaf_class=Text)

            start_time = time.time()
            for first_part in range(10):
                for second_part in range(10):
                    for third_part in range(10):
                        metadata_path = MetadataPath(f"{first_part:03}/"
                                                     f"{second_part:03}/"
                                                     f"{third_part:03}")
                        mtree.add_child_at(
                            Text(content=f"content of: {str(metadata_path)}"),
                            metadata_path)

            initialisation_duration = time.time() - start_time
            print(f"Initialised: {initialisation_duration:4f}")

            start_time = time.time()
            reference = mtree.write_out(metadata_store)
            write_out_duration = time.time() - start_time
            print(f"Written out: {write_out_duration:4f}")

            mtree = MTreeNode(leaf_class=Text,
                              realm=metadata_store,
                              reference=reference)
            start_time = time.time()
            mtree.read_in()
            read_in_duration = time.time() - start_time
            print(f"Read in: {read_in_duration:4f}")

            metadata_path = MetadataPath("5/5/xxx")
            start_time = time.time()
            mtree.add_child_at(
                Text(content=f"content of: {str(metadata_path)}"),
                metadata_path)
            add_duration = time.time() - start_time
            print(f"Added single entry: {add_duration:4f}")

            start_time = time.time()
            mtree.write_out()
            write_out_2nd_duration = time.time() - start_time
            print(f"Written out single entry: {write_out_2nd_duration:4f}")


if __name__ == '__main__':
    unittest.main()



import time
from typing import Dict, Tuple
from uuid import UUID

from dataladmetadatamodel.datasettree import DatasetTree
from dataladmetadatamodel.metadatarootrecord import MetadataRootRecord
from dataladmetadatamodel.versionlist import TreeVersionList


def create_tree_version_list_for_mrrs(
        mapper_family: str,
        realm: str,
        metadata_root_records: Dict[Tuple[UUID, str, str], MetadataRootRecord]
        ) -> TreeVersionList:

    dataset_tree = DatasetTree()

    for id_version_path, metadata_root_record in metadata_root_records.items():
        dataset_id, dataset_version, relative_path = id_version_path
        dataset_tree.add_dataset(relative_path, metadata_root_record)

    top_level_version = tuple(
        filter(
            lambda ivp_mrr: ivp_mrr[0][2] == "",
            metadata_root_records.items()))[0][0][1]

    tree_version_list = TreeVersionList()
    tree_version_list.set_dataset_tree(
        top_level_version,
        str(time.time()),
        dataset_tree
    )

    return tree_version_list
